from sqlalchemy import func

from models import Account
from database import DatabaseManager
from database.exceptions import OutOfAccountSpaceError, NonexistentAccountError, InsufficientBalance, InvalidOperation
from utils.constants import MIN_ACCOUNT_NUMBER, MAX_ACCOUNT_NUMBER


def get_next_id() -> int:
    with DatabaseManager.get_session() as session:
        current_max_id = session.query(func.max(Account.account_number)).scalar()
        current_max_id = current_max_id + 1 if current_max_id is not None else MIN_ACCOUNT_NUMBER

        if current_max_id > MAX_ACCOUNT_NUMBER:
            raise OutOfAccountSpaceError("Too many users already exist, cannot open new account")

        return current_max_id


def create_account() -> int:
    new_id = get_next_id()

    with DatabaseManager.get_session() as session:
        new_account = Account(account_number=new_id, balance=0)
        session.add(new_account)
        session.commit()
    return new_id


def get_account_balance(account_number: int) -> int:
    with DatabaseManager.get_session() as session:
        account: Account = session.query(Account).where(Account.account_number == account_number).one_or_none()
        if account is None:
            raise NonexistentAccountError(f"Account with number {account_number} doesn't exist")

        return account.balance


def withdraw_from_account(account_number: int, amount: int):
    modify_balance(account_number, amount, True)


def deposit_into_account(account_number: int, amount: int):
    modify_balance(account_number, amount, False)


def modify_balance(account_number: int, amount: int, subtract: bool):
    with DatabaseManager.get_session() as session:
        account: Account = session.query(Account).where(Account.account_number == account_number).one_or_none()
        if account is None:
            raise NonexistentAccountError(f"Account with number {account_number} doesn't exist")

        if subtract:
            account.balance += amount
        else:
            if account.balance - amount < 0:
                raise InsufficientBalance("Not enough funds on account to withdraw this much")
            account.balance -= amount

        session.commit()


def delete_account(account_number: int):
    with DatabaseManager.get_session() as session:
        account: Account = session.query(Account).where(Account.account_number == account_number).one_or_none()
        if account is None:
            raise NonexistentAccountError(f"Account with number {account_number} doesn't exist")

        if account.balance > 0:
            raise InvalidOperation("Cannot delete an account with leftover funds")

        session.delete(account)
        session.commit()
